// main controller for the installation
// TODO: playback of the recordings, automation, switch from open-loop to closed-loop with the openCV tracker
(
var imgPositions, imgIndexToAudioIndex, curPos, tarPos,
netAddress, localAddress, serialPort, serialListener,
moveTo, jogControl, jogHorizontal, jogVertical,
imgSelect, imgCalibrate, lastSelect,
trackerPos, trackerOffsetBaseDist, trackerOffset, trackLock,
dirTuples, dirAdjustRoutine, dirTuplesSeq;

"installation_audio_player.scd".loadRelative;

// init global vars
imgPositions = 9.collect({nil});
imgIndexToAudioIndex = [0, 1, -1, 2, -1, 3, -1, 4, 5];
curPos = Point.new(0, 0);
tarPos = Point.new(0, 0);
netAddress = NetAddr.new("127.0.0.1", 8080);
localAddress = NetAddr.new("127.0.0.1", 57120);
lastSelect = -1;
trackerOffsetBaseDist = 12000;
trackerOffset = [
	[trackerOffsetBaseDist * 1, trackerOffsetBaseDist * 1],
	[trackerOffsetBaseDist * 0, trackerOffsetBaseDist * 1],
	[trackerOffsetBaseDist * -1, trackerOffsetBaseDist * 1],

	[trackerOffsetBaseDist * 1, trackerOffsetBaseDist * 0],
	[trackerOffsetBaseDist * 0, trackerOffsetBaseDist * 0],
	[trackerOffsetBaseDist * -1, trackerOffsetBaseDist * 0],

	[trackerOffsetBaseDist * 1, trackerOffsetBaseDist * -1],
	[trackerOffsetBaseDist * 0, trackerOffsetBaseDist * -1],
	[trackerOffsetBaseDist * -1, trackerOffsetBaseDist * -1]
];
trackLock = true;
dirTuples = [ [ -1, -1 ], [ -1, 0 ], [ -1, 1 ], [ 0, -1 ], [ 0, 1 ], [ 1, -1 ], [ 1, 0 ], [ 1, 1 ] ];

~serialPort = SerialPort("/dev/ttyACM0", baudrate: 115200, crtscts: true);

// recieve motor feedback
~serialListener = Routine({
	var byte, str, res, valArray,
	stepper, limitSwitchNeg, limitSwitchPos, safeMode, limitPos;

	safeMode = false;

	loop{
		byte = ~serialPort.read;
		if(byte==13, {
			if(str[1].asString == "[", {
				valArray = str.asString.interpret.postln;
				netAddress.sendMsg("/STATE/SET", "{pos_message: \"x_pos: " + valArray[0] + " | y_pos: " + valArray[1] + "\"}");
				curPos = Point.new(valArray[0], valArray[1]);
				limitSwitchNeg = valArray[2];
				limitSwitchPos = valArray[3];
				if(safeMode && (limitSwitchNeg == limitSwitchPos), {
					safeMode = false;
					fork {
						netAddress.sendMsg("/STATE/SET", "{message: \"all clear\"}");
						netAddress.sendMsg("/STATE/SET", "{message_sub: \"all clear\"}");
						netAddress.sendMsg("/STATE/SET", "{message_public: \"all clear\"}");
						2.wait;
						netAddress.sendMsg("/STATE/SET", "{message: \"\"}");
						netAddress.sendMsg("/STATE/SET", "{message_sub: \"\"}");
						netAddress.sendMsg("/STATE/SET", "{message_public: \"\"}");
					}
				});
			}, {
				if(str[1..3].asString == "!!!", {
					netAddress.sendMsg("/STATE/SET", "{message: \"!!! limit switch still on after 1000 steps, this should NEVER happen\"}");
					netAddress.sendMsg("/STATE/SET", "{message_sub: \"!!! limit switch still on after 1000 steps, this should NEVER happen\"}");
					netAddress.sendMsg("/STATE/SET", "{message_public: \"!!! limit switch still on after 1000 steps, this should NEVER happen\"}");
				}, {
					safeMode = true;
					netAddress.sendMsg("/STATE/SET", "{message: \"!! limit hit, move the other direction\"}");
					netAddress.sendMsg("/STATE/SET", "{message_sub: \"!! limit hit, move the other direction\"}");
					netAddress.sendMsg("/STATE/SET", "{message_public: \"!! limit hit, move the other direction\"}");
				});
			});
			str = "";
		}, {str = str++byte.asAscii});
	};
}).play(AppClock);

// send new coordinates to the arduino / motors
moveTo = {arg point;
	if((point.x.abs < 8000) && (curPos.y.abs < 8000), {
		"move to".postln;
		point.postln;
		~serialPort.putAll(point.x.asInteger.asString ++ " " ++ point.y.asInteger.asString);
		~serialPort.put(10);
	});
};

jogControl = {arg axis;
	var jog, count = 0, jogRate= 0, jogDirection = 1;
	jog = Task({
		loop{
			count = (count + 0.01).clip(0, 1);
			//count.postln;
			jogRate = pow(count, 2) * 500;
			if(axis == '/jog_horizontal', {
				tarPos.x = curPos.x + (jogRate * jogDirection * -1);
			}, {
				tarPos.y = curPos.y + (jogRate * jogDirection);
			});
			//curPos.postln;
			moveTo.value(tarPos);
			0.1.wait
		};
	});
	OSCFunc({arg msg;
		msg.postln;
		if(msg[1] == 0, {count = 0; jogRate = 0; jog.pause()}, {jogDirection = msg[1]; jog.play(AppClock)});
	}, axis, netAddress);
};

jogHorizontal = jogControl.value('/jog_horizontal');
jogVertical = jogControl.value('/jog_vertical');

imgSelect = {
	OSCFunc({arg msg;
		var imgIndex, audioIndex;
		msg.postln;
		if((msg[1] > 0), {
			dirAdjustRoutine.stop;
			trackLock = false;
			imgIndex = msg[1] - 1;
			//imgIndex.postln;
			if(imgPositions[imgIndex] != nil, {tarPos = imgPositions[imgIndex].deepCopy; moveTo.value(tarPos)});
			9.do({arg i; if(imgIndex != i, {
				netAddress.sendMsg("/STATE/SET", "{img_" ++ (i + 1).asString ++ "_select: " ++ (i + 1).neg ++ "}")})});
			lastSelect = imgIndex;
			audioIndex = imgIndexToAudioIndex[imgIndex];
			//if(audioIndex != nil, {localAddress.sendMsg('/playTiling', audioIndex)});
			localAddress.sendMsg('/playTiling', audioIndex);
			dirTuplesSeq = 2.collect({dirTuples.deepCopy.scramble.collect({arg tup; [tup, tup * -1]}).flatten}).flatten;
			dirAdjustRoutine = Routine({
				120.wait;
				imgPositions[imgIndex] = curPos.deepCopy;
				trackLock = true;
				//10.wait;
				100.do({arg dirAdjust;
					"here".postln;
					if(dirAdjust < 8, {
						tarPos = imgPositions[imgIndex].deepCopy + (dirTuplesSeq[dirAdjust] * 200);
						moveTo.value(tarPos);
						3.wait;
						trackLock = false;
						15.wait;
						trackLock = true;
						//3.wait;
					}, {
						trackLock = false;
						15.wait;
						trackLock = true;
						15.wait;
					});
				});
				trackLock = false;
				//20.wait;
				//trackLock = true;
			}).play;
		}, {
			//lastSelect.postln;
			if(msg[1] == ((lastSelect + 1) * -1), {
				lastSelect = -1;
				dirAdjustRoutine.stop;
				trackLock = true;
			})
		});
	}, '/img_select', netAddress)
}.value;

imgCalibrate = {
	var calibrateHold, imgIndex, setPos;
	calibrateHold = Routine({
		var imageCalibrationIndex = 0;
		20.do({0.1.wait});
		[-1, 0, 1].do({arg row;
			[-1, 0, 1].do({arg col;
				imgPositions[imageCalibrationIndex] = setPos.deepCopy + Point.new(6700 * col, 6700 * row);
				imgPositions[imageCalibrationIndex].postln;
				imageCalibrationIndex = imageCalibrationIndex + 1;
			});
		});

		imgPositions.postln;
		netAddress.sendMsg("/STATE/SET", "{message: \"image calibrated\"}");
		netAddress.sendMsg("/STATE/SET", "{message_sub: \"image calibrated\"}");
		netAddress.sendMsg("/STATE/SET", "{message_public: \"image calibrated\"}");
	});

	OSCFunc({ arg msg;
		imgIndex = msg[1] - 1;
		if(imgIndex >= 0, {
			setPos = curPos.deepCopy;
			calibrateHold.play(AppClock);
		}, {
			calibrateHold.stop;
			calibrateHold.reset;
			netAddress.sendMsg("/STATE/SET", "{message: \"\"}");
			netAddress.sendMsg("/STATE/SET", "{message_sub: \"\"}");
			netAddress.sendMsg("/STATE/SET", "{message_public: \"\"}");
		});
	}, '/img_calibrate', netAddress);
}.value;

trackerPos = OSCFunc({arg msg;
	if((lastSelect.postln > -1) && trackLock.postln.not, {
		var distX, distY, isFocusing;
		distX = curPos.x - imgPositions[lastSelect].x;
		distY = curPos.y - imgPositions[lastSelect].y;
		isFocusing = false;
		if((imgPositions[lastSelect] != nil) && (distX.abs < 350) && ((msg[1] - trackerOffset[lastSelect][0]).abs > 3000), {
			tarPos.x = curPos.x + (10 * (msg[1] - trackerOffset[lastSelect][0]).sign);// * -1);
			isFocusing = true;
		});
		if((imgPositions[lastSelect] != nil) && (distY.abs < 350) && ((msg[2] - trackerOffset[lastSelect][1]).abs > 3000), {
			tarPos.y = curPos.y + (10 * (msg[2] - trackerOffset[lastSelect][1]).sign);// * -1);
			isFocusing = true;
		});
		if(isFocusing, {moveTo.value(tarPos)});
	})
}, '/trackerpos');
)

//in case of emergency
//~serialPort.close
//~serialPort = SerialPort.new("/dev/ttyACM0", baudrate: 115200, crtscts: true);
//~serialListener.reset
//~serialListener.play(AppClock);


